//******************************************************************************
// mystify.java:  Applet
//******************************************************************************

import java.applet.*;
import java.awt.*;

//=======================================================================
//  Mystify packet: m loops per n lines
//=======================================================================

class Packet
{
  private int lines;
  private int history;
  private Point[][] vert;
  private int top;    // first vertex in history
  private Point[] dir;  // direction
  private Color[] color;
  private int dR;
  private int dG;
  private int dB;     // color differences

  private int width;    // canvas size
  private int height;
  private Point ampl;   // random direction amplitude

  public Packet ( int l, int h, int wid, int hei ) {
    lines = l;
    history = h;
    vert = new Point[lines][history];
    width = wid;
    height = hei;
    dir = new Point[lines];
    ampl = new Point( width >> 6, height >> 6 );
    color = new Color[history];
    Generate();
    }

  private void Generate () {
    for ( int l = 0; l < lines; l++ ) {
      int x = (int)(Math.random()*width);
      int y = (int)(Math.random()*height);
      for ( int h = 0; h < history; h++ )
        vert[l][h] = new Point(x,y);
      }
    for ( int l = 0; l < lines; l++ )
      do {
        dir[l] = new Point( (int)(Math.random()*ampl.x*2)-ampl.x,
                  (int)(Math.random()*ampl.y*2)-ampl.y );
        } while ( dir[l].x == 0 && dir[l].y == 0 );
    for ( int h = 0; h < history; h++ )
      color[h] = Color.black;
    do {
      dR = (int)(Math.random()*3);
      dG = (int)(Math.random()*3);
      dB = (int)(Math.random()*3);
      } while ( dR == 0 && dG == 0 && dB == 0 );
    top = 0;
    }

  public void Step ( boolean col ) {
    int newtop = (top==history-1) ? 0 : top+1;
    for ( int l = 0; l < lines; l++ ) {
      int x = vert[l][top].x + dir[l].x;
      int y = vert[l][top].y + dir[l].y;
      while ( x < 0 || x >= width || y < 0 || y >= height ) {
          // bounce:
        do {
          dir[l].move( (int)(Math.random()*ampl.x*2)-ampl.x,
                 (int)(Math.random()*ampl.y*2)-ampl.y );
          } while ( dir[l].x == 0 && dir[l].y == 0 );
        x = vert[l][top].x + dir[l].x;
        y = vert[l][top].y + dir[l].y;
        }
      vert[l][newtop].move(x,y);
      }
    if ( col ) {
      int R = color[top].getRed() + dR;
      int G = color[top].getGreen() + dG;
      int B = color[top].getBlue() + dB;
      while ( R < 0 || R > 255 || G < 0 || G > 255 || B < 0 || B > 255 ) {
          // bounce:
        do {
          dR = (int)(Math.random()*6)-3;
          dG = (int)(Math.random()*6)-3;
          dB = (int)(Math.random()*6)-3;
          } while ( dR == 0 && dG == 0 && dB == 0 );
        R = color[top].getRed() + dR;
        G = color[top].getGreen() + dG;
        B = color[top].getBlue() + dB;
        }
      color[newtop] = new Color(R,G,B);
      }
    top = newtop;
    }

  private void DrawPolygon ( Graphics g, int h ) {
    int firstx = vert[0][h].x;
    int firsty = vert[0][h].y;
    int lastx = firstx;
    int lasty = firsty;
    for ( int l = 1; l < lines; l++ ) {
      int newx = vert[l][h].x;
      int newy = vert[l][h].y;
      g.drawLine(lastx,lasty,newx,newy);
      lastx = newx;
      lasty = newy;
      }
    g.drawLine(lastx,lasty,firstx,firsty);
    }

  public void Draw ( Graphics g ) {
    for ( int h = 0; h < history; h++ )
      if ( h != top+1 && (h != 0 || top != history-1) ) {
        g.setColor(color[h]);
        DrawPolygon(g,h);
        }
    }

  public void DrawTop ( Graphics g ) {
    g.setColor(color[top]);
    DrawPolygon(g,top);
    }

  public void EraseBottom ( Graphics g, Color bg ) {
    g.setColor(bg);
    DrawPolygon(g,(top==history-1)?0:top+1);
    }
}


//==============================================================================
// Main Class for applet mystify
//==============================================================================

public class mystify extends Applet implements Runnable
{
  boolean frozen = true;

  // THREAD SUPPORT:
  //    m_mystify is the Thread object for the applet
  //--------------------------------------------------------------------------
  Thread   m_mystify = null;

  // PARAMETER SUPPORT:
  //    Parameters allow an HTML author to pass information to the applet;
  // the HTML author specifies them using the <PARAM> tag within the <APPLET>
  // tag.  The following variables are used to store the values of the
  // parameters.
    //--------------------------------------------------------------------------

    // Members for applet parameters
    // <type>       <MemberVar>    = <Default Value>
    //--------------------------------------------------------------------------
  private int m_packets = 1;
  private int m_lines = 4;
  private int m_history = 4;
  private boolean m_colorChange = false;
  private int delay = 100;

    // Parameter names.  To change a name of a parameter, you need only make
  // a single change.  Simply modify the value of the parameter string below.
    //--------------------------------------------------------------------------
  private final String PARAM_packets = "packets";
  private final String PARAM_lines = "lines";
  private final String PARAM_history = "history";
  private final String PARAM_colorChange = "colorChange";
  private final String PARAM_delay = "delay";

  // APPLET INFO SUPPORT:
  //    The getAppletInfo() method returns a string describing the applet's
  // author, copyright date, or miscellaneous information.
    //--------------------------------------------------------------------------
  public String getAppletInfo()
  {
    return "Name: Mystify\r\n" +
           "Author: Pepca\r\n";
  }

  // PARAMETER SUPPORT
  //    The getParameterInfo() method returns an array of strings describing
  // the parameters understood by this applet.
  //
    // mystify Parameter Information:
    //  { "Name", "Type", "Description" },
    //--------------------------------------------------------------------------
  public String[][] getParameterInfo()
  {
    String[][] info =
    {
      { PARAM_packets, "int", "Number of mystify packets" },
      { PARAM_lines, "int", "Packet size" },
      { PARAM_history, "int", "Packet history" },
      { PARAM_colorChange, "boolean", "Change colors dynamically" },
      { PARAM_delay, "int", "Animation delay in miliseconds" },
    };
    return info;    
  }

  private Packet[] packets;   // []

    // backup store drawing:
  Dimension backupDim;
  Image backupImage;
  Graphics backupGraphics;

  // The init() method is called by the AWT when an applet is first loaded or
  // reloaded.  Override this method to perform whatever initialization your
  // applet needs, such as initializing data structures, loading images or
  // fonts, creating frame windows, setting the layout manager, or adding UI
  // components.
    //--------------------------------------------------------------------------
  public void init ()
  {
    // PARAMETER SUPPORT
    //    The following code retrieves the value of each parameter
    // specified with the <PARAM> tag and stores it in a member
    // variable.
    //----------------------------------------------------------------------
    String param;

    // packets: Number of mystify packets
    //----------------------------------------------------------------------
    param = getParameter(PARAM_packets);
    try {
      if (param != null)
          m_packets = Integer.parseInt(param);
      } catch ( Exception e ) {}

    // lines: Packet size
    //----------------------------------------------------------------------
    param = getParameter(PARAM_lines);
    try {
      if (param != null)
        m_lines = Integer.parseInt(param);
      } catch ( Exception e ) {}

    // history: Packet history
    //----------------------------------------------------------------------
    param = getParameter(PARAM_history);
    try {
      if (param != null)
        m_history = Integer.parseInt(param);
      } catch ( Exception e ) {}

    // colorChange: Change colors dynamically
    //----------------------------------------------------------------------
    param = getParameter(PARAM_colorChange);
    try {
      if (param != null)
        m_colorChange = Boolean.valueOf(param).booleanValue();
      } catch ( Exception e ) {}

    // delay: Animation delay
    //----------------------------------------------------------------------
    param = getParameter(PARAM_delay);
    try {
      if (param != null)
        delay = Integer.parseInt(param);
      } catch ( Exception e ) {}

    setBackground(new Color(255,255,255));
    
      // packets init:
    packets = new Packet[m_packets];
    Rectangle bounds = getBounds();
    for ( int p = 0; p < m_packets; p++ )
      packets[p] = new Packet(m_lines,m_history,bounds.width,bounds.height);
  }

  // mystify Update Handler
  //--------------------------------------------------------------------------
  public void update ( Graphics g )
  {
    Rectangle bounds = getBounds();
    Dimension d = new Dimension(bounds.width,bounds.height);

    if ( backupGraphics == null ||
       d.width != backupDim.width || d.height != backupDim.height ) {
      backupDim = d;
      backupImage = createImage(d.width,d.height);
      backupGraphics = backupImage.getGraphics();
      }
    backupGraphics.setColor(getBackground());
    backupGraphics.fillRect(0,0,d.width,d.height);
    backupGraphics.setColor(Color.black);
      //paint(g);
    for ( int p = 0; p < m_packets; p++ )
      packets[p].Draw(backupGraphics);
    g.drawImage(backupImage,0,0,this);
  }

  // mystify Paint Handler
  //--------------------------------------------------------------------------
  public void paint ( Graphics g )
  {
    for ( int p = 0; p < m_packets; p++ )
      packets[p].EraseBottom(g,getBackground());
    for ( int p = 0; p < m_packets; p++ )
      packets[p].Draw(g);
      //packets[p].DrawTop(g);
  }

  //    The start() method is called when the page containing the applet
  // first appears on the screen. The AppletWizard's initial implementation
  // of this method starts execution of the applet's thread.
  //--------------------------------------------------------------------------
  public void start ()
  {
    if ( !frozen ) {
      if ( m_mystify == null ) m_mystify = new Thread(this);
      m_mystify.start();
      }
  }
  
  //    The stop() method is called when the page containing the applet is
  // no longer on the screen. The AppletWizard's initial implementation of
  // this method stops execution of the applet's thread.
  //--------------------------------------------------------------------------
  public void stop ()
  {
    if ( m_mystify != null ) {
      m_mystify.stop();
      m_mystify = null;
      }
  }

  // THREAD SUPPORT
  //    The run() method is called when the applet's thread is started. If
  // your applet performs any ongoing activities without waiting for user
  // input, the code for implementing that behavior typically goes here. For
  // example, for an applet that performs animation, the run() method controls
  // the display of images.
  //--------------------------------------------------------------------------
  public void run ()
  {
        Thread.currentThread().setPriority(Thread.MIN_PRIORITY);

      //Remember the starting time:
        long startTime = System.currentTimeMillis();

    while ( Thread.currentThread() == m_mystify ) {
      try {
          // advance the animation:
        for ( int p = 0; p < m_packets; p++ )
          packets[p].Step(m_colorChange);
          // repaint:
        repaint();
          // animation delay:
        startTime += delay;
        long currentTime = System.currentTimeMillis();
        if ( currentTime < startTime ) Thread.sleep(startTime-currentTime);
        }
      catch ( InterruptedException e ) {
        break;
        }
      }
  }


  // MOUSE SUPPORT:
  //    The mouseEnter() method is called if the mouse cursor enters the
  // applet's portion of the screen.
  //--------------------------------------------------------------------------
  public boolean mouseEnter ( Event evt, int x, int y )
  {
    frozen = false;
    start();
    return true;
  }

  // MOUSE SUPPORT:
  //    The mouseExit() method is called if the mouse cursor leaves the
  // applet's portion of the screen.
  //--------------------------------------------------------------------------
  public boolean mouseExit ( Event evt, int x, int y )
  {
    frozen = true;
    stop();
    return true;
  }

}
