/// Quick-sort algorithm on <code>double[]</code> arrays.
class QuickSort {

  /// Minimal quick-sort array size (shorter segments are sorted by insert-sort).
  protected static final int QSMIN = 12;

  /**
   *  Sorting routine.
   *   @param  a Array which holds the sorted segment.
   *   @param  from Starting index of the segment.
   *   @param  len Length of the segment.
   */
  public static final void sort ( double[] a, int from, int len )
  {
    int i, j;
    int to = from + len;

    if ( len < QSMIN ) {               // short array => insert-sort
      while ( len > 1 ) {              // find minimum and swap it to a[from]
        double min = a[ j = from ];
        for ( i = from + 1; i < to; i++ )
          if ( a[i] < min ) {
            min = a[i];
            j = i;
          }
        if ( j != from ) {
          a[j] = a[from];
          a[from] = min;
        }
        from++;
        len--;
      }
      return;
    }

      // the array is long enough for quick-sort:
    double pivot = (a[from] + a[to-1]) * 0.5;
    i = from - 1; j = to;
    do {
      do i++;
      while ( i < to     && a[i] <= pivot );
      do j--;
      while ( j >= from  && a[j] >= pivot );
      if ( j < i ) break;
      double tmp = a[i];
      a[i] = a[j];
      a[j] = tmp;
    } while ( true );

      // j < i => sort [from to j] and [i to to-1]:
    if ( j - from + 1 < to - i ) {     // [from to j] first
      if ( j > from ) sort(a,from,j-from+1);
      sort(a,i,to-i);
    }
    else {                             // [i to to-1] first
      if ( to - i > 1 ) sort(a,i,to-i);
      sort(a,from,j-from+1);
    }
  }

}

/*
  Quick-sort in virtual memory (double[8M] array):
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,I5.0.1)         2.69s  100%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH)                3.12s  116%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.3.1IBM)      3.22s  120%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.4.1jrockit)  3.47s  129%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J1.4.2bHS)      3.95s  147%
BartonXP/2G (AsusA7V600,512M/333,IC35L040A,WXPH,J5.00.3802)     4.11s  153%
*/
